/*
 * pam_limits - impose resource limits when opening a user session
 *
 * 1.6 - modified for PLD (added process priority settings) 
 *       by Marcin Korzonek <mkorz@shadow.eu.org
 * 1.5 - Elliot Lee's "max system logins patch"
 * 1.4 - addressed bug in configuration file parser
 * 1.3 - modified the configuration file format
 * 1.2 - added 'debug' and 'conf=' arguments
 * 1.1 - added @group support    
 * 1.0 - initial release - Linux ONLY
 *
 * See end for Copyright information
 */

#if !(defined(linux))
#error THIS CODE IS KNOWN TO WORK ONLY ON LINUX !!!
#endif 

#include "../../_pam_aconf.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

#include <utmp.h>
#ifndef UT_USER  /* some systems have ut_name instead of ut_user */
#define UT_USER ut_user
#endif

#include <grp.h>
#include <pwd.h>

/* Module defines */
#define LIMITS_DEF_USER     0 /* limit was set by an user entry */
#define LIMITS_DEF_GROUP    1 /* limit was set by a group entry */
#define LIMITS_DEF_DEFAULT  2 /* limit was set by an default entry */
#define LIMITS_DEF_FORGROUP 3 /* limit set by &group */
#define LIMITS_DEF_FORALL   4 /* limit set by & */
#define LIMITS_DEF_NONE     5 /* this limit was not set yet */

const char *limits_def_names[] = {
	"USER",
	"GROUP",
	"DEFAULT",
	"FORGROUP",
	"FORALL",
	"NONE",
	NULL,
};

struct user_limits_struct {
    int src_soft;
    int src_hard;
    struct rlimit limit;
};

/* internal data */
struct pam_limit_s {
    int login_limit;     /* the max logins limit */
    int login_limit_def; /* which entry set the login limit */
    int flag_numsyslogins; /* whether to limit logins only for a
			      specific user or to count all logins */
    int priority;	 /* the priority to run user process with */
    struct user_limits_struct limits[RLIM_NLIMITS];
    char conf_file[BUFSIZ];
    char login_group[LINE_MAX+1];
};

#define LIMIT_LOGIN RLIM_NLIMITS+1
#define LIMIT_NUMSYSLOGINS RLIM_NLIMITS+2

#define LIMIT_PRI RLIM_NLIMITS+3

#define LIMIT_SOFT  1
#define LIMIT_HARD  2

#define PAM_SM_SESSION
#include "../../libpam/include/security/pam_modules.h"

#define PAM_GETPWNAM_R
#define PAM_GETGRNAM_R
#include "../../libpam/include/security/_pam_macros.h"

#ifdef HAVE_PWDBLIB
#include <pwdb/pwdb_map.h>
#else
#include <pwd.h>
#include <grp.h>
#endif

/* logging */
static void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("pam_limits", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

/* argument parsing */

#define PAM_DEBUG_ARG       0x0001
#define PAM_DO_SETREUID     0x0002

static int _pam_parse(int argc, const char **argv, struct pam_limit_s *pl)
{
     int ctrl=0;

     /* step through arguments */
     for (ctrl=0; argc-- > 0; ++argv) {

          /* generic options */
          if (!strcmp(*argv,"debug"))
                ctrl |= PAM_DEBUG_ARG;
          else if (!strncmp(*argv,"conf=",5))
                strncpy(pl->conf_file,*argv+5,sizeof(pl->conf_file) - 1);
	  else if (!strncmp(*argv,"change_uid",10))
		ctrl |= PAM_DO_SETREUID;
          else {
                _pam_log(LOG_ERR,"pam_parse: unknown option; %s",*argv);
          }
     }
     pl->conf_file[sizeof(pl->conf_file) - 1] = '\0';

     return ctrl;
}


/* limits stuff */
#ifdef DEFAULT_CONF_FILE
# define LIMITS_FILE DEFAULT_CONF_FILE
#else
# define LIMITS_FILE "/etc/security/limits.conf"
#endif

#define LIMIT_ERR 1 /* error setting a limit */
#define LOGIN_ERR 2 /* too many logins err */

static int is_on_group(const char *user_name, const char *group_name);

/* Counts the number of user logins and check against the limit*/
static int check_logins(const char *name, int limit, int ctrl,
			struct pam_limit_s *pl)
{
    struct utmp *ut;
    unsigned int count;

    if (ctrl & PAM_DEBUG_ARG) {
	 _pam_log(LOG_DEBUG, "checking logins for '%s' (maximum of %d)\n",
		  name, limit);
    }

    if (limit < 0)
        return 0; /* no limits imposed */
    if (limit == 0) /* maximum 0 logins ? */ {
        _pam_log(LOG_WARNING, "No logins allowed for '%s'\n", name);
        return LOGIN_ERR;
    }

    setutent();
    count = 0;
    while((ut = getutent())) {
#ifdef USER_PROCESS
        if (ut->ut_type != USER_PROCESS)
            continue;
#endif
        if (ut->UT_USER[0] == '\0')
            continue;
        if (!pl->flag_numsyslogins) {
	    if ((pl->login_limit_def == LIMITS_DEF_USER 
	         || pl->login_limit_def == LIMITS_DEF_FORGROUP
		 || pl->login_limit_def == LIMITS_DEF_FORALL)
	        && strncmp(name, ut->UT_USER, sizeof(ut->UT_USER)) != 0)
                continue;
	    if ((pl->login_limit_def == LIMITS_DEF_GROUP)
		&& !is_on_group(ut->UT_USER, pl->login_group))
                continue;
	}
        if (++count > limit)
            break;
    }
    endutent();
    if (count > limit) {
	if (name) {
	    _pam_log(LOG_WARNING, "Too many logins (max %d) for %s", limit,
		     name);
	} else {
	    _pam_log(LOG_WARNING, "Too many system logins (max %d)", limit);
	}
        return LOGIN_ERR;
    }
    return 0;
}

/* checks if a user is on a list of members of the GID 0 group */
static int is_on_list(char * const *list, const char *member)
{
    while (*list) {
        if (strcmp(*list, member) == 0)
            return 1;
        list++;
    }
    return 0;
}

/* Checks if a user is a member of a group, return non-zero if
   the user is in the group. */
static int is_on_group(const char *user_name, const char *group_name)
{
    struct passwd passwd, *pwd;
    struct group group, *grp;
    char uname[LINE_MAX], gname[LINE_MAX], *ubuf = NULL, *gbuf = NULL;
    size_t ubuflen, gbuflen;
    
    if (!user_name)
        return 0;
    if (!group_name)
        return 0;
    if (!strlen(user_name))
        return 0;
    if (!strlen(group_name))
        return 0;
    memset(uname, 0, sizeof(uname));
    strncpy(uname, user_name, sizeof(uname) - 1);
    memset(gname, 0, sizeof(gname));
    strncpy(gname, group_name, sizeof(gname) - 1);
        
    if (_pam_getpwnam_r(uname, &passwd, &ubuf, &ubuflen, &pwd) != 0)
	pwd = NULL;
    if (!pwd)
        return 0;

    /* the info about this group */
    if (_pam_getgrnam_r(gname, &group, &gbuf, &gbuflen, &grp) != 0)
	grp = NULL;
    if (!grp) {
        if (ubuf) {
            free(ubuf);
        }
        return 0;
    }
    
    /* first check: is a member of the group_name group ? */
    if (is_on_list(grp->gr_mem, uname)) {
        if (ubuf) {
            free(ubuf);
        }
        if (gbuf) {
            free(gbuf);
        }
        return 1;
    }

    /* next check: user primary group is group ? */
    if (pwd->pw_gid == grp->gr_gid) {
        if (ubuf) {
            free(ubuf);
        }
        if (gbuf) {
            free(gbuf);
        }
        return 1;
    }
        
    if (ubuf) {
        free(ubuf);
    }
    if (gbuf) {
        free(gbuf);
    }
    return 0;
}
    
static int init_limits(struct pam_limit_s *pl)
{
    int i;
    int retval = PAM_SUCCESS;

    D(("called."));

    /* If we were built on 2.4 and are running on 2.2, some of these will
     * give us EINVAL errors. */
    for(i = 0; i < RLIM_NLIMITS; i++) {
	int err = getrlimit(i, &pl->limits[i].limit);
	if(err != 0) {
            if(errno != EINVAL) {
		retval |= -1;
	    }
	}
    }
    
    pl->limits[RLIMIT_CPU].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_CPU].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_FSIZE].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_FSIZE].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_DATA].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_DATA].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_STACK].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_STACK].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_CORE].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_CORE].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_RSS].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_RSS].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_NPROC].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_NPROC].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_NOFILE].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_NOFILE].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_MEMLOCK].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_MEMLOCK].src_hard = LIMITS_DEF_NONE;

    pl->limits[RLIMIT_AS].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_AS].src_hard = LIMITS_DEF_NONE;

#ifdef RLIMIT_LOCKS
    pl->limits[RLIMIT_LOCKS].src_soft = LIMITS_DEF_NONE;
    pl->limits[RLIMIT_LOCKS].src_hard = LIMITS_DEF_NONE;
#endif

    pl->priority = 0;
    pl->login_limit = -2;
    pl->login_limit_def = LIMITS_DEF_NONE;

    return retval;
}    

static void process_limit(int source, const char *lim_type,
			  const char *lim_item, const char *lim_value,
			  int ctrl, struct pam_limit_s *pl)
{
    int limit_item;
    int limit_type = 0;
    rlim_t limit_value;
    const char **endptr = &lim_value;
    const char *value_orig = lim_value;
        
    /* avoid stupid warning about const manipulation */
    endptr = &lim_value;
    
    if (ctrl & PAM_DEBUG_ARG)
	 _pam_log(LOG_DEBUG, "%s: processing %s %s %s for %s\n",
		  __FUNCTION__,lim_type,lim_item,lim_value,
		  limits_def_names[source]);

    if (strcmp(lim_item, "cpu") == 0)
        limit_item = RLIMIT_CPU;
    else if (strcmp(lim_item, "fsize") == 0)
        limit_item = RLIMIT_FSIZE;
    else if (strcmp(lim_item, "data") == 0)
	limit_item = RLIMIT_DATA;
    else if (strcmp(lim_item, "stack") == 0)
	limit_item = RLIMIT_STACK;
    else if (strcmp(lim_item, "core") == 0)
	limit_item = RLIMIT_CORE;
    else if (strcmp(lim_item, "rss") == 0)
	limit_item = RLIMIT_RSS;
    else if (strcmp(lim_item, "nproc") == 0)
	limit_item = RLIMIT_NPROC;
    else if (strcmp(lim_item, "nofile") == 0)
	limit_item = RLIMIT_NOFILE;
    else if (strcmp(lim_item, "memlock") == 0)
	limit_item = RLIMIT_MEMLOCK;
    else if (strcmp(lim_item, "as") == 0)
	limit_item = RLIMIT_AS;
#ifdef RLIMIT_LOCKS
    else if (strcmp(lim_item, "locks") == 0)
	limit_item = RLIMIT_LOCKS;
#endif
    else if (strcmp(lim_item, "maxlogins") == 0) {
	limit_item = LIMIT_LOGIN;
	pl->flag_numsyslogins = 0;
    } else if (strcmp(lim_item, "maxsyslogins") == 0) {
	limit_item = LIMIT_NUMSYSLOGINS;
	pl->flag_numsyslogins = 1;
    } else if (strcmp(lim_item, "priority") == 0) {
	limit_item = LIMIT_PRI;
    } else {
        _pam_log(LOG_DEBUG,"unknown limit item '%s'", lim_item);
        return;
    }

    if (strcmp(lim_type,"soft")==0)
	limit_type=LIMIT_SOFT;
    else if (strcmp(lim_type, "hard")==0)
	limit_type=LIMIT_HARD;
    else if (strcmp(lim_type,"-")==0)
	limit_type=LIMIT_SOFT | LIMIT_HARD;
    else if (limit_item != LIMIT_LOGIN && limit_item != LIMIT_NUMSYSLOGINS) {
        _pam_log(LOG_DEBUG,"unknown limit type '%s'", lim_type);
        return;
    }
    
    if(strstr(lim_value,"unlimited")) {
       limit_value = RLIM_INFINITY;
    } else { 
#if (SIZEOF_RLIM_T > SIZEOF_UNSIGNED_LONG)
       limit_value = strtoull(lim_value, (char**)endptr, 10);
#else
       limit_value = strtoul(lim_value, (char**)endptr, 10);
#endif
       if (limit_value == 0 && value_orig == *endptr) { /* no chars read */
           if (strcmp(lim_value,"-") != 0) {
               _pam_log(LOG_DEBUG,"wrong limit value '%s'", lim_value);
               return;
           } else
               if (limit_item != LIMIT_LOGIN) {
                   if (ctrl & PAM_DEBUG_ARG)
                       _pam_log(LOG_DEBUG,
                               "'-' limit value valid for maxlogins type only");
                   return;
               } else
                   limit_value = -1;
       }

       /* one more special case when limiting logins */
       if ((source == LIMITS_DEF_FORALL || source == LIMITS_DEF_FORGROUP)
                && (limit_item != LIMIT_LOGIN)) {
           if (ctrl & PAM_DEBUG_ARG)
               _pam_log(LOG_DEBUG,
                        "'&' domain valid for maxlogins type only");
           return;
       }
                                                                                                                                                         
       switch(limit_item) {
           case RLIMIT_CPU:
               limit_value *= 60;
               break;
           case RLIMIT_FSIZE:
           case RLIMIT_DATA:
           case RLIMIT_STACK:
           case RLIMIT_CORE:
           case RLIMIT_RSS:
           case RLIMIT_MEMLOCK:
           case RLIMIT_AS:
               limit_value *= 1024;
               break;
       }
    }
    
    if ((limit_item != LIMIT_LOGIN) &&
	(limit_item != LIMIT_NUMSYSLOGINS) &&
	(limit_item != LIMIT_PRI)) {
        if (limit_type & LIMIT_SOFT) {
	    if (pl->limits[limit_item].src_soft < source) {
                return;
	    } else {
                pl->limits[limit_item].limit.rlim_cur = limit_value;
                pl->limits[limit_item].src_soft = source;
            }
	}
        if (limit_type & LIMIT_HARD) {
	    if (pl->limits[limit_item].src_hard < source) {
                return;
            } else {
                pl->limits[limit_item].limit.rlim_max = limit_value;
                pl->limits[limit_item].src_hard = source;
            }
	}
    } else 
	if (limit_item == LIMIT_PRI) {
		pl->priority = limit_value;
	} else {
	        if (pl->login_limit_def < source) {
	            return;
	        } else {
	            pl->login_limit = limit_value;
	            pl->login_limit_def = source;
        	}
	}
    return;
}

static int parse_config_file(const char *uname, int ctrl,
			     struct pam_limit_s *pl)
{
    FILE *fil;
    char buf[LINE_MAX];
    
#define CONF_FILE (pl->conf_file[0])?pl->conf_file:LIMITS_FILE
    /* check for the LIMITS_FILE */
    if (ctrl & PAM_DEBUG_ARG)
        _pam_log(LOG_DEBUG,"reading settings from '%s'", CONF_FILE);
    fil = fopen(CONF_FILE, "r");
    if (fil == NULL) {
        _pam_log (LOG_WARNING, "can not read settings from %s", CONF_FILE);
        return PAM_SERVICE_ERR;
    }
#undef CONF_FILE
    
    /* init things */
    memset(buf, 0, sizeof(buf));
    /* start the show */
    while (fgets(buf, LINE_MAX, fil) != NULL) {
        char domain[LINE_MAX];
        char ltype[LINE_MAX];
        char item[LINE_MAX];
        char value[LINE_MAX];
        int i,j;
        char *tptr;
        
        tptr = buf;
        /* skip the leading white space */
        while (*tptr && isspace(*tptr))
            tptr++;
        strncpy(buf, (const char *)tptr, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';
                                
        /* Rip off the comments */
        tptr = strchr(buf,'#');
        if (tptr)
            *tptr = '\0';
        /* Rip off the newline char */
        tptr = strchr(buf,'\n');
        if (tptr)
            *tptr = '\0';
        /* Anything left ? */
        if (!strlen(buf)) {
            memset(buf, 0, sizeof(buf));
            continue;
        }

        memset(domain, 0, sizeof(domain));
        memset(ltype, 0, sizeof(ltype));
        memset(item, 0, sizeof(item));
        memset(value, 0, sizeof(value));
        
        i = sscanf(buf,"%s%s%s%s", domain, ltype, item, value);
	D(("scanned line[%d]: domain[%s], ltype[%s], item[%s], value[%s]",
	   i, domain, ltype, item, value));

        for(j=0; j < strlen(domain); j++)
            domain[j]=tolower(domain[j]);
        for(j=0; j < strlen(ltype); j++)
            ltype[j]=tolower(ltype[j]);
        for(j=0; j < strlen(item); j++)
            item[j]=tolower(item[j]);
        for(j=0; j < strlen(value); j++)
            value[j]=tolower(value[j]);

        if (i == 4) { /* a complete line */
            if (strcmp(uname, domain) == 0) /* this user have a limit */
                process_limit(LIMITS_DEF_USER, ltype, item, value, ctrl, pl);
            else if (domain[0]=='@') {
	        _pam_log(LOG_DEBUG, "checking if %s is in group %s", uname, domain + 1);
                if (is_on_group(uname, domain+1)) {
		    strncpy(pl->login_group, domain+1, LINE_MAX);
                    process_limit(LIMITS_DEF_GROUP, ltype, item, value, ctrl,
				  pl);
		}
            } else if (domain[0]=='&') {
                if (domain[1]=='\0')
                    process_limit(LIMITS_DEF_FORALL, ltype, item, value, ctrl,
				  pl);
		 else if (is_on_group(uname, domain+1))
                    process_limit(LIMITS_DEF_FORGROUP, ltype, item, value, ctrl,
				  pl);
            } else if (strcmp(domain, "*") == 0)
                process_limit(LIMITS_DEF_DEFAULT, ltype, item, value, ctrl,
			      pl);
	} else if (i == 2 && ltype[0] == '-') { /* Probably a no-limit line */
	    if (strcmp(uname, domain) == 0) {
		_pam_log(LOG_DEBUG, "no limits for '%s'", uname);
		fclose(fil);
		return PAM_IGNORE;
	    } else if (domain[0] == '@' && is_on_group(uname, domain+1)) {
		_pam_log(LOG_DEBUG, "no limits for '%s' in group '%s'",
			 uname, domain+1);
		fclose(fil);
		return PAM_IGNORE;
	    }
        } else {
            _pam_log(LOG_DEBUG,"invalid line '%s' - skipped", buf);
	}
    }
    fclose(fil);
    return PAM_SUCCESS;    
}

static int setup_limits(const char * uname, uid_t uid,
			int ctrl, struct pam_limit_s *pl)
{
    int i;
    int retval = PAM_SUCCESS;
    
    /* do not impose limits on UID 0 accounts */
    if (uid == 0) {
        if (ctrl & PAM_DEBUG_ARG)
            _pam_log(LOG_DEBUG, "user '%s' has UID 0 - no limits imposed",
                                uname);
        if(pl->priority > 0)
            pl->priority = 0;
    } else
    for (i=0; i<RLIM_NLIMITS; i++) {
        if (pl->limits[i].limit.rlim_cur > pl->limits[i].limit.rlim_max)
            pl->limits[i].limit.rlim_cur = pl->limits[i].limit.rlim_max;
        retval |= setrlimit(i, &pl->limits[i].limit);
    }
    
    if (retval != PAM_SUCCESS)
        retval = LIMIT_ERR;

    i=setpriority(PRIO_PROCESS, 0, pl->priority);
    
    if (i != PAM_SUCCESS)
        retval = LIMIT_ERR;

    if ((uid != 0) && (pl->login_limit > 0)) {
        if (check_logins(uname, pl->login_limit, ctrl, pl) == LOGIN_ERR)
            retval |= LOGIN_ERR;
    } else if ((uid != 0) && (pl->login_limit == 0))
        retval |= LOGIN_ERR;
    return retval;
}
            
/* now the session stuff */
PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags,
                                   int argc, const char **argv)
{
    int retval;
    char *user_name;
    struct passwd passwd, *pwd;
    char *ubuf = NULL;
    size_t ubuflen;
    int ctrl;
    struct pam_limit_s pl;
    char error_buf[BUFSIZ];
    uid_t uid;

    D(("called."));

    memset(&pl, 0, sizeof(pl));

    ctrl = _pam_parse(argc, argv, &pl);
    retval = pam_get_item( pamh, PAM_USER, (void*) &user_name );
    if ( user_name == NULL || retval != PAM_SUCCESS ) {
        _pam_log(LOG_CRIT, "open_session - error recovering username");
        return PAM_SESSION_ERR;
     }
	
    if (_pam_getpwnam_r(user_name, &passwd, &ubuf, &ubuflen, &pwd) != 0)
	pwd = NULL;
    if (!pwd) {
        if (ctrl & PAM_DEBUG_ARG)
            _pam_log(LOG_WARNING, "open_session username '%s' does not exist",
                                   user_name);
        return PAM_SESSION_ERR;
    }
                     
    retval = init_limits(&pl);
    if (retval != PAM_SUCCESS) {
        _pam_log(LOG_WARNING, "cannot initialize");
	if (ubuf) {
            free(ubuf);
	}
        return PAM_IGNORE;
    }

    retval = parse_config_file(pwd->pw_name, ctrl, &pl);
    if (retval == PAM_IGNORE) {
	D(("the configuration file has an applicable '<domain> -' entry"));
	if (ubuf) {
            free(ubuf);
	}
	return PAM_SUCCESS;
    }
    if (retval != PAM_SUCCESS) {
        _pam_log(LOG_WARNING, "error parsing the configuration file");
	if (ubuf) {
            free(ubuf);
	}
        return PAM_IGNORE;
    }

    if (ctrl & PAM_DO_SETREUID)
	setreuid(pwd->pw_uid, -1);

    retval = setup_limits(pwd->pw_name, pwd->pw_uid, ctrl, &pl);
    if (retval & LOGIN_ERR) {
	snprintf(error_buf, sizeof(error_buf), "Too many logins for '%s'.\n",
		 pwd->pw_name);
#if 0
	if (pam_get_item(pamh, PAM_CONV, (const void **)&conv) == PAM_SUCCESS)
	    conv->conv(1, messages, &responses, conv->appdata_ptr);
	else
#endif
            printf("%s", error_buf);
        sleep(2);
        if (ubuf) {
            free(ubuf);
        }
        return PAM_PERM_DENIED;
    }

    if (ubuf) {
        free(ubuf);
    }
    return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags,
                                    int argc, const char **argv)
{
     /* nothing to do */
     return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_limits_modstruct = {
     "pam_limits",
     NULL,
     NULL,
     NULL,
     pam_sm_open_session,
     pam_sm_close_session,
     NULL
};
#endif

/*
 * Copyright (c) Cristian Gafton, 1996-1997, <gafton@redhat.com>
 *                                              All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 * 
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
